@@ -7,8 +7,6 @@ import torch.nn.functional as F
 from math import sqrt
 from einops import rearrange, repeat, reduce
 
-from axial_positional_embedding import AxialPositionalEmbedding
-
 import alphafold2_pytorch.constants as constants
 from alphafold2_pytorch.reversible import ReversibleSequence
 
@@ -339,6 +337,8 @@ class Alphafold2(nn.Module):
         pos_token = 3,
         num_tokens = constants.NUM_AMINO_ACIDS,
         num_embedds = constants.NUM_EMBEDDS_TR,
+        max_num_msas = constants.MAX_NUM_MSA,
+        max_num_templates = constants.MAX_NUM_TEMPLATES,
         attn_dropout = 0.,
         ff_dropout = 0.,
         reversible = False,
@@ -351,17 +351,20 @@ class Alphafold2(nn.Module):
         layers_sparse_attn = cast_tuple(sparse_self_attn, depth)
 
         self.token_emb = nn.Embedding(num_tokens, dim)
-        self.pos_emb = AxialPositionalEmbedding(dim, (max_seq_len, max_seq_len))
+        self.pos_emb = nn.Embedding(max_seq_len, dim)
+        self.pos_emb_ax = nn.Embedding(max_seq_len, dim)
 
         # multiple sequence alignment position embedding
 
         self.msa_pos_emb = nn.Embedding(max_seq_len, dim)
-        self.msa_num_pos_emb = nn.Embedding(constants.MAX_NUM_MSA, dim)
+        self.msa_num_pos_emb = nn.Embedding(max_num_msas, dim)
 
         # template embedding
 
-        self.template_emb = nn.Embedding(constants.DISTOGRAM_BUCKETS, dim)
-        self.template_pos_emb = AxialPositionalEmbedding(dim, (max_seq_len, max_seq_len))
+        self.template_dist_emb = nn.Embedding(constants.DISTOGRAM_BUCKETS, dim)
+        self.template_num_pos_emb = nn.Embedding(max_num_templates, dim)
+        self.template_pos_emb = nn.Embedding(max_seq_len, dim)
+        self.template_pos_emb_ax = nn.Embedding(max_seq_len, dim)
 
         # custom embedding projection
 
@@ -423,11 +426,13 @@ class Alphafold2(nn.Module):
         msa = None,
         mask = None,
         msa_mask = None,
-        templates = None,
+        templates_seq = None,
+        templates_dist = None,
         templates_mask = None,
         embedds = None,
     ):
         n, device = seq.shape[1], seq.device
+        n_range = torch.arange(n, device = device)
 
         # unpack (AA_code, atom_pos)
 
@@ -451,7 +456,8 @@ class Alphafold2(nn.Module):
 
         # axial positional embedding
 
-        x += self.pos_emb(x)
+        pos_emb = rearrange(self.pos_emb(n_range), 'i d -> () i () d') + rearrange(self.pos_emb_ax(n_range), 'j d -> () () j d')
+        x += rearrange(pos_emb, 'b i j d -> b (i j) d')
 
         # embed multiple sequence alignment (msa)
 
@@ -474,19 +480,34 @@ class Alphafold2(nn.Module):
 
         # trunk (template attention, wip)
 
-        if exists(templates):
-            _, num_templates, *_ = templates.shape
+        if exists(templates_seq):
+            assert exists(templates_dist), 'template binned residue distances must be supplied `templates_dist`'
+            _, num_templates, *_ = templates_seq.shape
 
             # embed template
 
-            t = self.template_emb(templates)
+            t_seq = self.token_emb(templates_seq)
+            t_dist = self.template_dist_emb(templates_dist)
+
+            t_seq = rearrange(t_seq, 'b t i d -> b t i () d') + rearrange(t_seq, 'b t j d -> b t () j d')
+            t = t_seq + t_dist
+
             template_shape = rearrange(t, 'b t h w d -> (b t) h w d').shape
 
+            # template pos emb
+
+            template_num_pos_emb = self.template_num_pos_emb(torch.arange(num_templates, device = device))
+
+            t += rearrange(template_num_pos_emb, 't d-> () t () () d')
             t = rearrange(t, 'b t h w d -> (b t) (h w) d')
-            t += self.template_pos_emb(t)
+
+            pos_emb = rearrange(self.template_pos_emb(n_range), 'i d -> () i () d') + rearrange(self.template_pos_emb_ax(n_range), 'j d -> () () j d')
+            pos_emb = rearrange(pos_emb, 'b i j d -> b (i j) d')
+            t += pos_emb
 
             if exists(templates_mask):
-                t_mask = rearrange(templates_mask, 'b t h w -> (b t) (h w)')
+                t_mask = rearrange(templates_mask, 'b t i -> b t i ()') * rearrange(templates_mask, 'b t j -> b t () j')
+                t_mask = rearrange(t_mask, 'b t h w -> (b t) (h w)')
 
             assert t.shape[-2:] == x.shape[-2:]
 
@@ -499,7 +520,7 @@ class Alphafold2(nn.Module):
                 # concat template and primary sequence, for axial attention along the template dimension
                 # very similar to the attention across the time axis from https://arxiv.org/abs/2102.05095
 
-                x = rearrange(x, 'b (h w) d -> (b h w) () d', h = n)
+                x = rearrange(x, 'b n d -> (b n) () d', n = n ** 2)
                 t = rearrange(t, '(b t) n d -> (b n) t d', t = num_templates)
 
                 y_mask = None
@@ -511,7 +532,7 @@ class Alphafold2(nn.Module):
                 y = torch.cat((x, t), dim = 1)
                 y = cross_attn(y, mask = y_mask) + y
 
-                x = rearrange(y[:, 0], '(b h w) d -> b (h w) d', h = n, w = n)
+                x = rearrange(y[:, 0], '(b n) d -> b n d', n = n ** 2)
                 t = rearrange(y[:, 1:], '(b n) t d -> (b t) n d', n = (n ** 2))
 
                 t = ff(t) + t
